#!/usr/bin/env python3
"""
Example usage of the GARCH(1,1) with GED innovations model.

This script demonstrates how to use the garch11_ged package for volatility modeling.
"""

import numpy as np
import matplotlib.pyplot as plt
import sys
import os

# Add the project root to the Python path for development
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

try:
    import garch11_ged
except ImportError:
    print("Error: garch11_ged module not found. Please build the package first.")
    print("Run: pip install .")
    sys.exit(1)


def generate_sample_data(n=1000, mu=0.0, omega=0.0001, alpha=0.1, beta=0.8, nu=1.5):
    """
    Generate synthetic GARCH(1,1) data with GED innovations.
    
    Parameters:
    -----------
    n : int
        Number of observations
    mu : float
        Mean return
    omega : float
        GARCH constant term
    alpha : float
        ARCH parameter
    beta : float
        GARCH parameter
    nu : float
        GED shape parameter
    """
    print(f"Generating {n} observations of synthetic GARCH(1,1)-GED data...")
    print(f"True parameters: mu={mu}, omega={omega}, alpha={alpha}, beta={beta}, nu={nu}")
    
    np.random.seed(42)
    
    # Initialize arrays
    returns = np.zeros(n)
    h = np.zeros(n)  # conditional variance
    
    # Initial variance
    h[0] = omega / (1 - alpha - beta)  # unconditional variance
    
    # Generate GED distributed innovations (approximation using normal)
    # In practice, we would use the exact GED distribution
    for t in range(n):
        # GED innovation (approximated with normal for simplicity)
        z_t = np.random.normal(0, 1)
        epsilon_t = np.sqrt(h[t]) * z_t
        returns[t] = mu + epsilon_t
        
        # Update conditional variance for next period
        if t < n - 1:
            h[t+1] = omega + alpha * epsilon_t**2 + beta * h[t]
    
    return returns


def main():
    """Main example function."""
    print("=" * 60)
    print("GARCH(1,1) with GED Innovations - Example Usage")
    print("=" * 60)
    
    # 1. Generate sample data
    n_obs = 1000
    true_params = {
        'mu': 0.001,
        'omega': 0.0001, 
        'alpha': 0.1,
        'beta': 0.8,
        'nu': 1.5
    }
    
    returns = generate_sample_data(n_obs, **true_params)
    
    print(f"\nSample statistics:")
    print(f"  Mean: {np.mean(returns):.6f}")
    print(f"  Std:  {np.std(returns):.6f}")
    print(f"  Min:  {np.min(returns):.6f}")
    print(f"  Max:  {np.max(returns):.6f}")
    
    # 2. Fit GARCH model
    print("\n" + "="*40)
    print("Fitting GARCH(1,1)-GED Model")
    print("="*40)
    
    model = garch11_ged.GARCH11_GED()
    
    print("Fitting model...")
    success = model.fit(returns, max_iter=1000, tolerance=1e-6)
    
    if success:
        print("✓ Model fitted successfully!")
    else:
        print("⚠ Model fitting may not have converged.")
    
    # 3. Display results
    print("\n" + model.summary())
    
    # 4. Compare with true parameters
    estimated_params = model.get_parameters()
    param_names = ['mu', 'omega', 'alpha', 'beta', 'nu']
    
    print("\nParameter Comparison:")
    print("-" * 50)
    print(f"{'Parameter':<10} {'True':<12} {'Estimated':<12} {'Diff':<10}")
    print("-" * 50)
    
    for i, name in enumerate(param_names):
        true_val = true_params[name]
        est_val = estimated_params[i]
        diff = abs(est_val - true_val)
        print(f"{name:<10} {true_val:<12.6f} {est_val:<12.6f} {diff:<10.6f}")
    
    # 5. Get fitted results
    volatilities = model.get_volatilities()
    residuals = model.get_residuals()
    std_residuals = model.get_standardized_residuals()
    
    print(f"\nModel Diagnostics:")
    print(f"  Log-Likelihood: {model.get_log_likelihood():.4f}")
    print(f"  AIC:           {model.get_aic():.4f}")
    print(f"  BIC:           {model.get_bic():.4f}")
    
    # 6. Volatility forecasting
    print("\n" + "="*40)
    print("Volatility Forecasting")
    print("="*40)
    
    forecast_horizon = 10
    forecasts = model.predict_volatility(forecast_horizon)
    
    print(f"\n{forecast_horizon}-step ahead volatility forecasts:")
    for i, vol_forecast in enumerate(forecasts, 1):
        print(f"  h+{i}: {vol_forecast:.6f}")
    
    # 7. Plot results (if matplotlib is available)
    try:
        plot_results(returns, volatilities, residuals, std_residuals, forecasts)
    except ImportError:
        print("\nMatplotlib not available. Skipping plots.")
    
    # 8. Performance comparison
    print("\n" + "="*40)
    print("Performance Test")
    print("="*40)
    
    performance_test()
    
    print("\n" + "="*60)
    print("Example completed successfully!")
    print("="*60)


def plot_results(returns, volatilities, residuals, std_residuals, forecasts):
    """Plot the results."""
    print("\nGenerating plots...")
    
    fig, axes = plt.subplots(2, 2, figsize=(15, 10))
    fig.suptitle('GARCH(1,1)-GED Model Results', fontsize=16)
    
    # Plot 1: Returns and volatility
    t = np.arange(len(returns))
    axes[0, 0].plot(t, returns, alpha=0.7, label='Returns', linewidth=0.8)
    axes[0, 0].plot(t, volatilities, 'r-', label='Conditional Volatility', linewidth=1.5)
    axes[0, 0].plot(t, -volatilities, 'r-', linewidth=1.5)
    axes[0, 0].set_title('Returns and Conditional Volatility')
    axes[0, 0].set_xlabel('Time')
    axes[0, 0].set_ylabel('Value')
    axes[0, 0].legend()
    axes[0, 0].grid(True, alpha=0.3)
    
    # Plot 2: Residuals
    axes[0, 1].plot(residuals, alpha=0.7)
    axes[0, 1].set_title('Residuals')
    axes[0, 1].set_xlabel('Time')
    axes[0, 1].set_ylabel('Residuals')
    axes[0, 1].grid(True, alpha=0.3)
    
    # Plot 3: Standardized residuals
    axes[1, 0].plot(std_residuals, alpha=0.7)
    axes[1, 0].set_title('Standardized Residuals')
    axes[1, 0].set_xlabel('Time')
    axes[1, 0].set_ylabel('Standardized Residuals')
    axes[1, 0].grid(True, alpha=0.3)
    
    # Plot 4: Volatility forecasts
    forecast_t = np.arange(len(volatilities), len(volatilities) + len(forecasts))
    axes[1, 1].plot(t[-50:], volatilities[-50:], 'b-', label='Historical Volatility')
    axes[1, 1].plot(forecast_t, forecasts, 'r-', marker='o', label='Forecasts')
    axes[1, 1].set_title('Volatility Forecasts')
    axes[1, 1].set_xlabel('Time')
    axes[1, 1].set_ylabel('Volatility')
    axes[1, 1].legend()
    axes[1, 1].grid(True, alpha=0.3)
    
    plt.tight_layout()
    plt.show()


def performance_test():
    """Test the performance of the implementation."""
    import time
    
    # Generate larger dataset for performance testing
    print("Testing performance with larger dataset...")
    large_returns = np.random.normal(0, 0.01, 5000)
    
    # Time the fitting process
    model = garch11_ged.GARCH11_GED()
    
    start_time = time.time()
    model.fit(large_returns)
    end_time = time.time()
    
    fitting_time = end_time - start_time
    print(f"  Fitted 5000 observations in {fitting_time:.3f} seconds")
    print(f"  Performance: {len(large_returns)/fitting_time:.0f} observations/second")
    
    # Test forecasting performance
    start_time = time.time()
    forecasts = model.predict_volatility(100)
    end_time = time.time()
    
    forecast_time = end_time - start_time
    print(f"  Generated 100-step forecast in {forecast_time:.6f} seconds")


def convenience_function_example():
    """Demonstrate the convenience function."""
    print("\n" + "="*40)
    print("Convenience Function Example")
    print("="*40)
    
    # Generate sample data
    returns = np.random.normal(0, 0.01, 500)
    
    # Fit model in one line
    model = garch11_ged.fit_garch(returns)
    
    print("Model fitted using convenience function:")
    print(f"  Parameters: {model.get_parameters()}")
    print(f"  Log-likelihood: {model.get_log_likelihood():.4f}")


if __name__ == "__main__":
    main()
    convenience_function_example() 