#pragma once

#include <vector>
#include <memory>
#include <string>
#include <stdexcept>

namespace garch {

/**
 * @brief GARCH(1,1) model with GED (Generalized Error Distribution) innovations
 * 
 * This class implements a GARCH(1,1) volatility model with GED distributed innovations.
 * The model is defined as:
 * 
 * Mean equation: r_t = mu + epsilon_t
 * Variance equation: sigma^2_t = omega + alpha * epsilon^2_{t-1} + beta * sigma^2_{t-1}
 * Innovation distribution: epsilon_t ~ GED(0, sigma_t, nu)
 */
class GARCH11_GED {
public:
    /**
     * @brief Default constructor
     */
    GARCH11_GED();
    
    /**
     * @brief Constructor with parameters
     * @param mu Mean parameter
     * @param omega Constant term in variance equation
     * @param alpha ARCH parameter
     * @param beta GARCH parameter
     * @param nu GED shape parameter
     */
    GARCH11_GED(double mu, double omega, double alpha, double beta, double nu);
    
    /**
     * @brief Destructor
     */
    ~GARCH11_GED();
    
    /**
     * @brief Fit the model to data
     * @param returns Vector of return data
     * @param max_iter Maximum number of iterations
     * @param tolerance Convergence tolerance
     * @return True if convergence achieved
     */
    bool fit(const std::vector<double>& returns, int max_iter = 1000, double tolerance = 1e-6);
    
    /**
     * @brief Predict volatility for h periods ahead
     * @param horizon Prediction horizon
     * @return Vector of predicted volatilities
     */
    std::vector<double> predict_volatility(int horizon = 1) const;
    
    /**
     * @brief Get fitted conditional volatilities
     * @return Vector of fitted volatilities
     */
    std::vector<double> get_volatilities() const;
    
    /**
     * @brief Get fitted residuals
     * @return Vector of fitted residuals
     */
    std::vector<double> get_residuals() const;
    
    /**
     * @brief Get standardized residuals
     * @return Vector of standardized residuals
     */
    std::vector<double> get_standardized_residuals() const;
    
    /**
     * @brief Get log-likelihood value
     * @return Log-likelihood
     */
    double get_log_likelihood() const;
    
    /**
     * @brief Get AIC (Akaike Information Criterion)
     * @return AIC value
     */
    double get_aic() const;
    
    /**
     * @brief Get BIC (Bayesian Information Criterion)
     * @return BIC value
     */
    double get_bic() const;
    
    /**
     * @brief Get parameter estimates
     * @return Vector containing [mu, omega, alpha, beta, nu]
     */
    std::vector<double> get_parameters() const;
    
    /**
     * @brief Set parameters manually
     * @param mu Mean parameter
     * @param omega Constant term
     * @param alpha ARCH parameter
     * @param beta GARCH parameter
     * @param nu GED shape parameter
     */
    void set_parameters(double mu, double omega, double alpha, double beta, double nu);
    
    /**
     * @brief Check if model is fitted
     * @return True if fitted
     */
    bool is_fitted() const;
    
    /**
     * @brief Get model summary as string
     * @return Summary string
     */
    std::string summary() const;

private:
    // Model parameters
    double mu_;      // Mean
    double omega_;   // Constant term
    double alpha_;   // ARCH parameter
    double beta_;    // GARCH parameter
    double nu_;      // GED shape parameter
    
    // Data and fitted values
    std::vector<double> returns_;
    std::vector<double> volatilities_;
    std::vector<double> residuals_;
    std::vector<double> standardized_residuals_;
    
    // Model state
    bool fitted_;
    double log_likelihood_;
    size_t n_obs_;
    
    // Private methods
    double ged_density(double x, double nu) const;
    double ged_log_density(double x, double nu) const;
    double gamma_function(double x) const;
    std::vector<double> compute_initial_parameters(const std::vector<double>& returns) const;
    double compute_log_likelihood(const std::vector<double>& params, const std::vector<double>& returns) const;
    void filter_volatilities(const std::vector<double>& returns);
    bool optimize_parameters(const std::vector<double>& returns, int max_iter, double tolerance);
    void validate_parameters() const;
    std::vector<double> numerical_gradient(const std::vector<double>& params, const std::vector<double>& returns) const;
};

} // namespace garch 