#include "../include/garch11_ged.hpp"
#include <cmath>
#include <algorithm>
#include <numeric>
#include <sstream>
#include <iomanip>

namespace garch {

// Mathematical constants
const double PI = 3.14159265358979323846;
const double EPS = 1e-12;

// Gamma function implementation using Lanczos approximation
double GARCH11_GED::gamma_function(double x) const {
    if (x < 0.5) {
        return PI / (std::sin(PI * x) * gamma_function(1.0 - x));
    }
    
    x -= 1.0;
    const double p[] = {
        0.99999999999980993,
        676.5203681218851,
        -1259.1392167224028,
        771.32342877765313,
        -176.61502916214059,
        12.507343278686905,
        -0.13857109526572012,
        9.9843695780195716e-6,
        1.5056327351493116e-7
    };
    
    double result = p[0];
    for (int i = 1; i < 9; ++i) {
        result += p[i] / (x + i);
    }
    
    double t = x + 7.5;
    return std::sqrt(2.0 * PI) * std::pow(t, x + 0.5) * std::exp(-t) * result;
}

// GED density function (standardized)
double GARCH11_GED::ged_density(double x, double nu) const {
    if (nu <= 0) {
        throw std::invalid_argument("GED shape parameter nu must be positive");
    }
    
    double lambda = std::sqrt(std::pow(0.5, 2.0/nu) * gamma_function(1.0/nu) / gamma_function(3.0/nu));
    double g = nu / (lambda * std::pow(2.0, 1.0 + 1.0/nu) * gamma_function(1.0/nu));
    
    return g * std::exp(-0.5 * std::pow(std::abs(x/lambda), nu));
}

// GED log-density function
double GARCH11_GED::ged_log_density(double x, double nu) const {
    if (nu <= 0) {
        throw std::invalid_argument("GED shape parameter nu must be positive");
    }
    
    double lambda = std::sqrt(std::pow(0.5, 2.0/nu) * gamma_function(1.0/nu) / gamma_function(3.0/nu));
    double log_g = std::log(nu) - std::log(lambda) - (1.0 + 1.0/nu) * std::log(2.0) - std::log(gamma_function(1.0/nu));
    
    return log_g - 0.5 * std::pow(std::abs(x/lambda), nu);
}

// Default constructor
GARCH11_GED::GARCH11_GED() 
    : mu_(0.0), omega_(0.01), alpha_(0.1), beta_(0.8), nu_(2.0), 
      fitted_(false), log_likelihood_(0.0), n_obs_(0) {
}

// Constructor with parameters
GARCH11_GED::GARCH11_GED(double mu, double omega, double alpha, double beta, double nu)
    : mu_(mu), omega_(omega), alpha_(alpha), beta_(beta), nu_(nu),
      fitted_(false), log_likelihood_(0.0), n_obs_(0) {
    validate_parameters();
}

// Destructor
GARCH11_GED::~GARCH11_GED() = default;

// Validate parameters
void GARCH11_GED::validate_parameters() const {
    if (omega_ <= 0) {
        throw std::invalid_argument("omega must be positive");
    }
    if (alpha_ < 0 || beta_ < 0) {
        throw std::invalid_argument("alpha and beta must be non-negative");
    }
    if (alpha_ + beta_ >= 1.0) {
        throw std::invalid_argument("alpha + beta must be less than 1 for stationarity");
    }
    if (nu_ <= 0) {
        throw std::invalid_argument("nu must be positive");
    }
}

// Compute initial parameter estimates
std::vector<double> GARCH11_GED::compute_initial_parameters(const std::vector<double>& returns) const {
    if (returns.empty()) {
        throw std::invalid_argument("Returns vector cannot be empty");
    }
    
    // Compute sample mean
    double mean = std::accumulate(returns.begin(), returns.end(), 0.0) / returns.size();
    
    // Compute sample variance
    double variance = 0.0;
    for (double r : returns) {
        variance += (r - mean) * (r - mean);
    }
    variance /= (returns.size() - 1);
    
    // Initial parameter estimates
    std::vector<double> params(5);
    params[0] = mean;                    // mu
    params[1] = 0.1 * variance;          // omega
    params[2] = 0.1;                     // alpha
    params[3] = 0.8;                     // beta
    params[4] = 2.0;                     // nu (GED shape)
    
    return params;
}

// Filter volatilities using GARCH(1,1) recursion
void GARCH11_GED::filter_volatilities(const std::vector<double>& returns) {
    n_obs_ = returns.size();
    volatilities_.resize(n_obs_);
    residuals_.resize(n_obs_);
    standardized_residuals_.resize(n_obs_);
    
    // Initial volatility (sample standard deviation)
    double sum_sq = 0.0;
    for (size_t i = 0; i < 10 && i < returns.size(); ++i) {
        double r = returns[i] - mu_;
        sum_sq += r * r;
    }
    double h_init = sum_sq / std::min(10.0, static_cast<double>(returns.size()));
    
    // GARCH(1,1) filtering
    volatilities_[0] = std::sqrt(h_init);
    residuals_[0] = returns[0] - mu_;
    standardized_residuals_[0] = residuals_[0] / volatilities_[0];
    
    for (size_t t = 1; t < n_obs_; ++t) {
        // GARCH(1,1): h_t = omega + alpha * eps_{t-1}^2 + beta * h_{t-1}^2
        double h_t = omega_ + alpha_ * residuals_[t-1] * residuals_[t-1] + 
                     beta_ * volatilities_[t-1] * volatilities_[t-1];
        
        volatilities_[t] = std::sqrt(std::max(h_t, EPS));
        residuals_[t] = returns[t] - mu_;
        standardized_residuals_[t] = residuals_[t] / volatilities_[t];
    }
}

// Compute log-likelihood
double GARCH11_GED::compute_log_likelihood(const std::vector<double>& params, 
                                           const std::vector<double>& returns) const {
    // Temporarily set parameters
    double mu = params[0];
    double omega = params[1];
    double alpha = params[2];
    double beta = params[3];
    double nu = params[4];
    
    // Parameter bounds checking
    if (omega <= 0 || alpha < 0 || beta < 0 || alpha + beta >= 1.0 || nu <= 0) {
        return -1e10; // Return very negative log-likelihood for invalid parameters
    }
    
    size_t n = returns.size();
    std::vector<double> h(n);
    
    // Initial volatility
    double sum_sq = 0.0;
    for (size_t i = 0; i < 10 && i < n; ++i) {
        double r = returns[i] - mu;
        sum_sq += r * r;
    }
    h[0] = sum_sq / std::min(10.0, static_cast<double>(n));
    
    double log_likelihood = 0.0;
    
    // First observation
    double z0 = (returns[0] - mu) / std::sqrt(h[0]);
    log_likelihood += ged_log_density(z0, nu) - 0.5 * std::log(h[0]);
    
    // Remaining observations
    for (size_t t = 1; t < n; ++t) {
        double eps_prev = returns[t-1] - mu;
        h[t] = omega + alpha * eps_prev * eps_prev + beta * h[t-1];
        h[t] = std::max(h[t], EPS);
        
        double z_t = (returns[t] - mu) / std::sqrt(h[t]);
        log_likelihood += ged_log_density(z_t, nu) - 0.5 * std::log(h[t]);
    }
    
    return log_likelihood;
}

// Numerical gradient computation
std::vector<double> GARCH11_GED::numerical_gradient(const std::vector<double>& params, 
                                                    const std::vector<double>& returns) const {
    const double h = 1e-6;
    std::vector<double> grad(params.size());
    
    for (size_t i = 0; i < params.size(); ++i) {
        std::vector<double> params_plus = params;
        std::vector<double> params_minus = params;
        
        params_plus[i] += h;
        params_minus[i] -= h;
        
        double ll_plus = compute_log_likelihood(params_plus, returns);
        double ll_minus = compute_log_likelihood(params_minus, returns);
        
        grad[i] = (ll_plus - ll_minus) / (2.0 * h);
    }
    
    return grad;
}

// Optimize parameters using simple gradient ascent
bool GARCH11_GED::optimize_parameters(const std::vector<double>& returns, 
                                      int max_iter, double tolerance) {
    std::vector<double> params = compute_initial_parameters(returns);
    double learning_rate = 0.001;
    double prev_ll = compute_log_likelihood(params, returns);
    
    for (int iter = 0; iter < max_iter; ++iter) {
        std::vector<double> grad = numerical_gradient(params, returns);
        
        // Update parameters
        for (size_t i = 0; i < params.size(); ++i) {
            params[i] += learning_rate * grad[i];
        }
        
        // Ensure parameter constraints
        params[1] = std::max(params[1], 1e-6);  // omega > 0
        params[2] = std::max(params[2], 0.0);   // alpha >= 0
        params[3] = std::max(params[3], 0.0);   // beta >= 0
        params[4] = std::max(params[4], 0.1);   // nu > 0
        
        // Ensure stationarity
        if (params[2] + params[3] >= 0.999) {
            double sum = params[2] + params[3];
            params[2] *= 0.999 / sum;
            params[3] *= 0.999 / sum;
        }
        
        double current_ll = compute_log_likelihood(params, returns);
        
        if (std::abs(current_ll - prev_ll) < tolerance) {
            break;
        }
        
        // Adaptive learning rate
        if (current_ll < prev_ll) {
            learning_rate *= 0.5;
        } else {
            learning_rate = std::min(learning_rate * 1.01, 0.01);
        }
        
        prev_ll = current_ll;
    }
    
    // Set final parameters
    mu_ = params[0];
    omega_ = params[1];
    alpha_ = params[2];
    beta_ = params[3];
    nu_ = params[4];
    log_likelihood_ = prev_ll;
    
    return true;
}

// Fit the model
bool GARCH11_GED::fit(const std::vector<double>& returns, int max_iter, double tolerance) {
    if (returns.size() < 10) {
        throw std::invalid_argument("Need at least 10 observations to fit the model");
    }
    
    returns_ = returns;
    
    // Optimize parameters
    bool converged = optimize_parameters(returns, max_iter, tolerance);
    
    // Filter volatilities with optimized parameters
    filter_volatilities(returns);
    
    fitted_ = true;
    return converged;
}

// Predict volatility
std::vector<double> GARCH11_GED::predict_volatility(int horizon) const {
    if (!fitted_) {
        throw std::runtime_error("Model must be fitted before prediction");
    }
    
    std::vector<double> predictions(horizon);
    
    // Get last volatility and residual
    double h_last = volatilities_.back() * volatilities_.back();
    double eps_last = residuals_.back();
    
    // Unconditional variance for long-term forecast
    double uncond_var = omega_ / (1.0 - alpha_ - beta_);
    
    for (int i = 0; i < horizon; ++i) {
        if (i == 0) {
            // One-step ahead: h_{t+1} = omega + alpha * eps_t^2 + beta * h_t
            predictions[i] = std::sqrt(omega_ + alpha_ * eps_last * eps_last + beta_ * h_last);
        } else {
            // Multi-step ahead: converges to unconditional volatility
            double weight = std::pow(alpha_ + beta_, i);
            double h_forecast = uncond_var * (1.0 - weight) + h_last * weight;
            predictions[i] = std::sqrt(h_forecast);
        }
    }
    
    return predictions;
}

// Getter methods
std::vector<double> GARCH11_GED::get_volatilities() const {
    if (!fitted_) {
        throw std::runtime_error("Model must be fitted first");
    }
    return volatilities_;
}

std::vector<double> GARCH11_GED::get_residuals() const {
    if (!fitted_) {
        throw std::runtime_error("Model must be fitted first");
    }
    return residuals_;
}

std::vector<double> GARCH11_GED::get_standardized_residuals() const {
    if (!fitted_) {
        throw std::runtime_error("Model must be fitted first");
    }
    return standardized_residuals_;
}

double GARCH11_GED::get_log_likelihood() const {
    if (!fitted_) {
        throw std::runtime_error("Model must be fitted first");
    }
    return log_likelihood_;
}

double GARCH11_GED::get_aic() const {
    if (!fitted_) {
        throw std::runtime_error("Model must be fitted first");
    }
    return -2.0 * log_likelihood_ + 2.0 * 5.0; // 5 parameters
}

double GARCH11_GED::get_bic() const {
    if (!fitted_) {
        throw std::runtime_error("Model must be fitted first");
    }
    return -2.0 * log_likelihood_ + std::log(n_obs_) * 5.0; // 5 parameters
}

std::vector<double> GARCH11_GED::get_parameters() const {
    return {mu_, omega_, alpha_, beta_, nu_};
}

void GARCH11_GED::set_parameters(double mu, double omega, double alpha, double beta, double nu) {
    mu_ = mu;
    omega_ = omega;
    alpha_ = alpha;
    beta_ = beta;
    nu_ = nu;
    validate_parameters();
}

bool GARCH11_GED::is_fitted() const {
    return fitted_;
}

std::string GARCH11_GED::summary() const {
    if (!fitted_) {
        return "Model not fitted yet.";
    }
    
    std::ostringstream oss;
    oss << std::fixed << std::setprecision(6);
    
    oss << "GARCH(1,1) with GED innovations\n";
    oss << "==============================\n\n";
    oss << "Parameters:\n";
    oss << "  mu    = " << mu_ << "\n";
    oss << "  omega = " << omega_ << "\n";
    oss << "  alpha = " << alpha_ << "\n";
    oss << "  beta  = " << beta_ << "\n";
    oss << "  nu    = " << nu_ << "\n\n";
    oss << "Model Statistics:\n";
    oss << "  Log-Likelihood = " << log_likelihood_ << "\n";
    oss << "  AIC            = " << get_aic() << "\n";
    oss << "  BIC            = " << get_bic() << "\n";
    oss << "  Observations   = " << n_obs_ << "\n";
    
    return oss.str();
}

} // namespace garch 