#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/numpy.h>
#include "../include/garch11_ged.hpp"

namespace py = pybind11;

PYBIND11_MODULE(garch11_ged, m) {
    m.doc() = "GARCH(1,1) with GED innovations - A fast C++ implementation";

    py::class_<garch::GARCH11_GED>(m, "GARCH11_GED")
        .def(py::init<>(), "Default constructor")
        .def(py::init<double, double, double, double, double>(),
             "Constructor with parameters",
             py::arg("mu"), py::arg("omega"), py::arg("alpha"), py::arg("beta"), py::arg("nu"))
        
        .def("fit", &garch::GARCH11_GED::fit,
             "Fit the GARCH model to data",
             py::arg("returns"), py::arg("max_iter") = 1000, py::arg("tolerance") = 1e-6,
             R"pbdoc(
                Fit the GARCH(1,1)-GED model to return data.
                
                Parameters
                ----------
                returns : list of float
                    Time series of returns
                max_iter : int, optional
                    Maximum number of optimization iterations (default: 1000)
                tolerance : float, optional
                    Convergence tolerance (default: 1e-6)
                
                Returns
                -------
                bool
                    True if convergence achieved
             )pbdoc")
        
        .def("predict_volatility", &garch::GARCH11_GED::predict_volatility,
             "Predict future volatility",
             py::arg("horizon") = 1,
             R"pbdoc(
                Predict volatility for h periods ahead.
                
                Parameters
                ----------
                horizon : int, optional
                    Number of periods to forecast (default: 1)
                
                Returns
                -------
                list of float
                    Predicted volatilities
             )pbdoc")
        
        .def("get_volatilities", &garch::GARCH11_GED::get_volatilities,
             "Get fitted conditional volatilities")
        
        .def("get_residuals", &garch::GARCH11_GED::get_residuals,
             "Get fitted residuals")
        
        .def("get_standardized_residuals", &garch::GARCH11_GED::get_standardized_residuals,
             "Get standardized residuals")
        
        .def("get_log_likelihood", &garch::GARCH11_GED::get_log_likelihood,
             "Get log-likelihood value")
        
        .def("get_aic", &garch::GARCH11_GED::get_aic,
             "Get AIC (Akaike Information Criterion)")
        
        .def("get_bic", &garch::GARCH11_GED::get_bic,
             "Get BIC (Bayesian Information Criterion)")
        
        .def("get_parameters", &garch::GARCH11_GED::get_parameters,
             "Get parameter estimates as [mu, omega, alpha, beta, nu]")
        
        .def("set_parameters", &garch::GARCH11_GED::set_parameters,
             "Set parameters manually",
             py::arg("mu"), py::arg("omega"), py::arg("alpha"), py::arg("beta"), py::arg("nu"))
        
        .def("is_fitted", &garch::GARCH11_GED::is_fitted,
             "Check if model is fitted")
        
        .def("summary", &garch::GARCH11_GED::summary,
             "Get model summary as string")
        
        // Properties for easier access
        .def_property_readonly("parameters", &garch::GARCH11_GED::get_parameters,
                              "Model parameters as [mu, omega, alpha, beta, nu]")
        .def_property_readonly("volatilities", &garch::GARCH11_GED::get_volatilities,
                              "Fitted conditional volatilities")
        .def_property_readonly("residuals", &garch::GARCH11_GED::get_residuals,
                              "Fitted residuals")
        .def_property_readonly("standardized_residuals", &garch::GARCH11_GED::get_standardized_residuals,
                              "Standardized residuals")
        .def_property_readonly("log_likelihood", &garch::GARCH11_GED::get_log_likelihood,
                              "Log-likelihood value")
        .def_property_readonly("aic", &garch::GARCH11_GED::get_aic,
                              "AIC value")
        .def_property_readonly("bic", &garch::GARCH11_GED::get_bic,
                              "BIC value")
        .def_property_readonly("fitted", &garch::GARCH11_GED::is_fitted,
                              "Whether model is fitted")
        
        .def("__repr__", [](const garch::GARCH11_GED &g) {
            return g.is_fitted() ? g.summary() : "GARCH11_GED(not fitted)";
        });

    // Module-level convenience functions
    m.def("fit_garch", [](const std::vector<double>& returns, int max_iter = 1000, double tolerance = 1e-6) {
        garch::GARCH11_GED model;
        model.fit(returns, max_iter, tolerance);
        return model;
    }, "Convenience function to fit GARCH model in one call",
       py::arg("returns"), py::arg("max_iter") = 1000, py::arg("tolerance") = 1e-6);
} 