import pytest
import numpy as np
import sys
import os

# Add the project root to the Python path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

try:
    import garch11_ged
except ImportError:
    # If the module is not built yet, skip these tests
    pytest.skip("garch11_ged module not built", allow_module_level=True)


class TestGARCH11GED:
    
    def setup_method(self):
        """Setup test data before each test method."""
        np.random.seed(42)
        # Generate sample return data with GARCH-like properties
        n = 1000
        self.returns = np.random.normal(0, 0.01, n)
        
        # Simple synthetic GARCH data
        h = np.zeros(n)
        omega, alpha, beta = 0.0001, 0.1, 0.8
        h[0] = 0.01**2
        
        for t in range(1, n):
            h[t] = omega + alpha * self.returns[t-1]**2 + beta * h[t-1]
            self.returns[t] = np.sqrt(h[t]) * np.random.normal(0, 1)
    
    def test_constructor_default(self):
        """Test default constructor."""
        model = garch11_ged.GARCH11_GED()
        assert not model.is_fitted()
        params = model.get_parameters()
        assert len(params) == 5
        assert params[1] > 0  # omega > 0
        assert params[2] >= 0  # alpha >= 0
        assert params[3] >= 0  # beta >= 0
        assert params[4] > 0  # nu > 0
    
    def test_constructor_with_parameters(self):
        """Test constructor with parameters."""
        mu, omega, alpha, beta, nu = 0.01, 0.0001, 0.1, 0.8, 2.0
        model = garch11_ged.GARCH11_GED(mu, omega, alpha, beta, nu)
        
        params = model.get_parameters()
        assert params[0] == mu
        assert params[1] == omega
        assert params[2] == alpha
        assert params[3] == beta
        assert params[4] == nu
    
    def test_parameter_validation(self):
        """Test parameter validation."""
        # Test invalid omega (negative)
        with pytest.raises(Exception):
            garch11_ged.GARCH11_GED(0, -0.01, 0.1, 0.8, 2.0)
        
        # Test invalid alpha (negative)
        with pytest.raises(Exception):
            garch11_ged.GARCH11_GED(0, 0.01, -0.1, 0.8, 2.0)
        
        # Test invalid stationarity condition (alpha + beta >= 1)
        with pytest.raises(Exception):
            garch11_ged.GARCH11_GED(0, 0.01, 0.6, 0.5, 2.0)
        
        # Test invalid nu (negative)
        with pytest.raises(Exception):
            garch11_ged.GARCH11_GED(0, 0.01, 0.1, 0.8, -1.0)
    
    def test_fit_basic(self):
        """Test basic model fitting."""
        model = garch11_ged.GARCH11_GED()
        result = model.fit(self.returns)
        
        assert result is True  # Should converge
        assert model.is_fitted()
        
        # Check that we get reasonable parameter estimates
        params = model.get_parameters()
        assert len(params) == 5
        assert params[1] > 0  # omega > 0
        assert params[2] >= 0  # alpha >= 0
        assert params[3] >= 0  # beta >= 0
        assert params[2] + params[3] < 1  # stationarity
        assert params[4] > 0  # nu > 0
    
    def test_fit_insufficient_data(self):
        """Test fitting with insufficient data."""
        model = garch11_ged.GARCH11_GED()
        short_returns = [0.01, 0.02, 0.01]  # Only 3 observations
        
        with pytest.raises(Exception):
            model.fit(short_returns)
    
    def test_get_results_before_fitting(self):
        """Test that getting results before fitting raises errors."""
        model = garch11_ged.GARCH11_GED()
        
        with pytest.raises(Exception):
            model.get_volatilities()
        
        with pytest.raises(Exception):
            model.get_residuals()
        
        with pytest.raises(Exception):
            model.get_log_likelihood()
    
    def test_get_results_after_fitting(self):
        """Test getting results after fitting."""
        model = garch11_ged.GARCH11_GED()
        model.fit(self.returns)
        
        # Test volatilities
        volatilities = model.get_volatilities()
        assert len(volatilities) == len(self.returns)
        assert all(v > 0 for v in volatilities)
        
        # Test residuals
        residuals = model.get_residuals()
        assert len(residuals) == len(self.returns)
        
        # Test standardized residuals
        std_residuals = model.get_standardized_residuals()
        assert len(std_residuals) == len(self.returns)
        
        # Test log-likelihood
        ll = model.get_log_likelihood()
        assert isinstance(ll, float)
        
        # Test AIC and BIC
        aic = model.get_aic()
        bic = model.get_bic()
        assert isinstance(aic, float)
        assert isinstance(bic, float)
        assert bic > aic  # BIC has stronger penalty
    
    def test_predict_volatility(self):
        """Test volatility prediction."""
        model = garch11_ged.GARCH11_GED()
        model.fit(self.returns)
        
        # Test single-step prediction
        forecast_1 = model.predict_volatility(1)
        assert len(forecast_1) == 1
        assert forecast_1[0] > 0
        
        # Test multi-step prediction
        forecast_10 = model.predict_volatility(10)
        assert len(forecast_10) == 10
        assert all(f > 0 for f in forecast_10)
    
    def test_predict_before_fitting(self):
        """Test that prediction before fitting raises error."""
        model = garch11_ged.GARCH11_GED()
        
        with pytest.raises(Exception):
            model.predict_volatility(1)
    
    def test_summary(self):
        """Test model summary."""
        model = garch11_ged.GARCH11_GED()
        
        # Summary before fitting
        summary_before = model.summary()
        assert "not fitted" in summary_before.lower()
        
        # Summary after fitting
        model.fit(self.returns)
        summary_after = model.summary()
        assert "garch" in summary_after.lower()
        assert "parameters" in summary_after.lower()
        assert "log-likelihood" in summary_after.lower()
    
    def test_properties(self):
        """Test property access."""
        model = garch11_ged.GARCH11_GED()
        model.fit(self.returns)
        
        # Test properties
        assert len(model.parameters) == 5
        assert len(model.volatilities) == len(self.returns)
        assert len(model.residuals) == len(self.returns)
        assert len(model.standardized_residuals) == len(self.returns)
        assert isinstance(model.log_likelihood, float)
        assert isinstance(model.aic, float)
        assert isinstance(model.bic, float)
        assert model.fitted is True
    
    def test_set_parameters(self):
        """Test manual parameter setting."""
        model = garch11_ged.GARCH11_GED()
        
        mu, omega, alpha, beta, nu = 0.001, 0.0001, 0.05, 0.9, 1.5
        model.set_parameters(mu, omega, alpha, beta, nu)
        
        params = model.get_parameters()
        assert params[0] == mu
        assert params[1] == omega
        assert params[2] == alpha
        assert params[3] == beta
        assert params[4] == nu
    
    def test_convenience_function(self):
        """Test convenience function for fitting."""
        model = garch11_ged.fit_garch(self.returns)
        assert model.is_fitted()
        assert len(model.get_parameters()) == 5
    
    def test_repr(self):
        """Test string representation."""
        model = garch11_ged.GARCH11_GED()
        
        # Before fitting
        repr_before = repr(model)
        assert "not fitted" in repr_before.lower()
        
        # After fitting
        model.fit(self.returns)
        repr_after = repr(model)
        assert "garch" in repr_after.lower()


class TestEdgeCases:
    
    def test_constant_returns(self):
        """Test with constant returns."""
        constant_returns = [0.01] * 100
        model = garch11_ged.GARCH11_GED()
        
        # This should still work but may have convergence issues
        try:
            model.fit(constant_returns)
            assert model.is_fitted()
        except Exception:
            # Convergence failure is acceptable for constant data
            pass
    
    def test_very_small_returns(self):
        """Test with very small returns."""
        small_returns = np.random.normal(0, 1e-8, 100)
        model = garch11_ged.GARCH11_GED()
        
        try:
            model.fit(small_returns)
            assert model.is_fitted()
        except Exception:
            # Numerical issues are acceptable for very small data
            pass
    
    def test_large_returns(self):
        """Test with large returns."""
        large_returns = np.random.normal(0, 10, 100)
        model = garch11_ged.GARCH11_GED()
        
        result = model.fit(large_returns)
        assert model.is_fitted()


if __name__ == "__main__":
    pytest.main([__file__]) 